/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggTheora SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE Theora SOURCE CODE IS COPYRIGHT (C) 2002-2007                *
 * by the Xiph.Org Foundation and contributors http://www.xiph.org/ *
 *                                                                  *
 ********************************************************************

  function: example dumpvid application; dumps  Theora streams
  last mod: $Id: dump_video.c 15400 2008-10-15 12:10:58Z tterribe $

 ********************************************************************/

/* By Robin Watts (robin@wss.co.uk). Actually, this was originally
 * dump_video.c by Mauricio Piacentini (mauricio at xiph.org), but I've
 * hacked it around horribly, so it's probably not nice to blame him
 * for it in any way! */

/* Enable/disable the following to determine if we post process */
#undef POST_PROCESS


#if !defined(_GNU_SOURCE)
#define _GNU_SOURCE
#endif
#if !defined(_LARGEFILE_SOURCE)
#define _LARGEFILE_SOURCE
#endif
#if !defined(_LARGEFILE64_SOURCE)
#define _LARGEFILE64_SOURCE
#endif
#if !defined(_FILE_OFFSET_BITS)
#define _FILE_OFFSET_BITS 64
#endif

#include <stdio.h>
//#if !defined(_WIN32)
//#include <getopt.h>
//#include <unistd.h>
//#else
//#include "getopt.h"
//#endif
#include <stdlib.h>
#include <string.h>
//#include <sys/timeb.h>
//#include <sys/types.h>
//#include <sys/stat.h>
/*Yes, yes, we're going to hell.*/
//#if defined(_WIN32)
//#include <io.h>
//#endif
//#include <fcntl.h>
#include <math.h>
//#include <signal.h>
#include "theora/theora.h"
#include "tremolo/ivorbiscodec.h"
#include "tremolo/codec_internal.h"

#ifdef _WIN32_WCE
#include <windows.h>

#define clock GetTickCount
#define CLOCKS_PER_SEC 1000
typedef long clock_t;

#else
#include <time.h>
#endif

char text[4096];

void Output(const char *fmt, ...)
{
#ifdef _WIN32_WCE
  va_list  ap;
  char    *t = text;
  WCHAR    uni[4096];
  WCHAR   *u = uni;

  va_start(ap,fmt);
  vsprintf(text, fmt, ap);
  va_end(ap);

  while (*t != 0)
  {
      *u++ = (WCHAR)(*t++);
  }
  *u++ = 0;
  OutputDebugString(uni);
#else
  vfprintf(stderr, fmt, ap);
#endif
}

#ifdef _WIN32_WCE
typedef struct _RawFrameBufferInfo
{
   WORD wFormat;
   WORD wBPP;
   VOID *pFramePointer;
   int  cxStride;
   int  cyStride;
   int  cxPixels;
   int  cyPixels;
}
RawFrameBufferInfo;

#define GETRAWFRAMEBUFFER   0x00020001
#define FORMAT_565 1
#define FORMAT_555 2
#define FORMAT_OTHER 3

struct GXDisplayProperties
{
    DWORD cxWidth;
    DWORD cyHeight; /* notice lack of 'th' in the word height */
    long cbxPitch;  /* number of bytes to move right one x pixel - can be negative */
    long cbyPitch;  /* number of bytes to move down one y pixel - can be negative */
    long cBPP;      /* # of bits in each pixel */
    DWORD ffFormat; /* format flags */
};
typedef struct GXDisplayProperties(*getGXProperties)(void);

void *find_screen_mem(int *width,
                      int *height)
{
    HDC                hdc;
    int                result;
    RawFrameBufferInfo frameBufferInfo;

    /* First, we'll try using GETRAWFRAMEBUFFER. This should work on all
     * modern WinCEs, but some 2003's get it wrong. */
    hdc = GetDC(NULL);
    result = ExtEscape(hdc, GETRAWFRAMEBUFFER, 0, NULL,
                       sizeof(RawFrameBufferInfo), (char *)&frameBufferInfo);

    if (result > 0)
    {
        *width  = frameBufferInfo.cxPixels;
        *height = frameBufferInfo.cyPixels;
        return frameBufferInfo.pFramePointer;
    }

    return NULL;
}
#endif

extern void yuv420_2_rgb565(short         *rgb_frame,
                            unsigned char *y,
                            unsigned char *u,
                            unsigned char *v,
                            int            height,
                            int            width,
                            int            y_stride,
                            int            uv_stride,
                            int            dither);

static int screen_width  = 0;
static int screen_height = 0;
static int dither = 0;

void *open_screen(int width, int height)
{
#ifdef _WIN32_WCE
  short *buff;

  buff = find_screen_mem(&screen_width, &screen_height);

  Output("Screen %d x %d @ %x\n", screen_width, screen_height, buff);

  if (buff != NULL)
  {
    memset(buff, 0, screen_width*screen_height*2);

    /* Centre the output */
    if (width < screen_width)
    {
      buff += ((screen_width-width)>>1);
    }
    if (height < screen_height)
    {
      buff += ((screen_height-height)>>1)*screen_width;
    }

    return buff;
  }

#endif
  return NULL;
}

/* Helper; just grab some more compressed bitstream and sync it for
   page extraction */
int buffer_data(FILE *in,ogg_sync_state *oy){
  char *buffer=ogg_sync_bufferin(oy,4096);
  int bytes=fread(buffer,1,4096,in);
  ogg_sync_wrote(oy,bytes);
  return(bytes);
}

/* never forget that globals are a one-way ticket to Hell */
/* Ogg and codec state for demux/decode */
ogg_sync_state    oy;
ogg_page          og;
ogg_stream_state  vo;
ogg_stream_state  to;
theora_info       ti;
theora_comment    tc;
theora_state      td;
vorbis_info       vi;
vorbis_dsp_state  vd;
vorbis_comment    vc;

int              theora_p;
int              vorbis_p;
int              stateflag;

/* single frame video buffering */
int          videobuf_ready;
ogg_int64_t  videobuf_granulepos;
double       videobuf_time;

/* single audio fragment audio buffering */
int          audiobuf_ready;
ogg_int16_t *audiobuf;
ogg_int64_t  audiobuf_granulepos; /* time position of last sample */
#define PCM_SIZE (1<<15)
ogg_int16_t  pcm[PCM_SIZE];
int          pcmstart; /* In samples */
int          pcmsent;  /* In samples */
int          pcmend;   /* In samples */
ogg_int64_t  audio_samples;
int          audio_sendtime;

typedef enum
{
    Playback_DecodeOnly = 0,
    Playback_Synced     = 1,
    Playback_Unsynced   = 2
} Playback;

Playback        playback;

#ifdef _WIN32_WCE
HWAVEOUT wince_audio_dev;
WAVEHDR whdr1;
WAVEHDR whdr2;
#endif

void open_audio(void)
{
#ifdef _WIN32_WCE
  WAVEFORMATEX wfx;
  MMRESULT     res;

  wince_audio_dev = NULL;
  wfx.wFormatTag      = WAVE_FORMAT_PCM;
  wfx.nChannels       = vi.channels;
  wfx.nSamplesPerSec  = vi.rate;
  wfx.wBitsPerSample  = 16;
  wfx.nBlockAlign     = (wfx.wBitsPerSample*vi.channels)>>3;
  wfx.nAvgBytesPerSec = vi.rate*wfx.nBlockAlign;
  wfx.cbSize          = 0;
  res = waveOutOpen(&wince_audio_dev,
                    0,
                    &wfx,
                    0,
                    0,
                    CALLBACK_NULL);
  if (res!= MMSYSERR_NOERROR)
  {
    /* Failed */
    wince_audio_dev = NULL;
  }
  whdr1.dwBufferLength = 0;
  whdr1.dwFlags = WHDR_DONE;
  whdr2.dwBufferLength = 0;
  whdr2.dwFlags = WHDR_DONE;
  pcmstart = 0;
  pcmsent  = 0;
  pcmend   = 0;
  audio_samples = 0;
  audio_sendtime = clock();
#endif
}

#if 1
#define AUDIO_CHECK()
#else
#define AUDIO_CHECK() do {audio_check();}while(0==1)
void audio_check(void)
{
    if ((pcmstart <0) || (pcmstart >= PCM_SIZE))
    {
        Output("pcmstart out of range! %04x\n", pcmstart);
        pcmsent = (pcmsent>>1)<<1;
    }
    if ((pcmsent <0) || (pcmsent >= PCM_SIZE))
    {
        Output("pcmsent out of range! %04x\n", pcmsent);
        pcmsent = (pcmsent>>1)<<1;
    }
    if ((pcmend <0) || (pcmend >= PCM_SIZE))
    {
        Output("pcmend out of range! %04x\n", pcmend);
        pcmend = (pcmend>>1)<<1;
    }
    //Output("start=%04x sent=%04x end=%04x\n", pcmstart, pcmsent, pcmend);
    if (pcmstart == pcmend)
    {
        if (pcmsent != pcmstart)
        {
            Output("pcmsent Broken!\n");
            pcmsent = (pcmsent>>1)<<1;
        }
    }
    if (pcmstart < pcmend)
    {
        if ((pcmsent < pcmstart) || (pcmsent > pcmend))
        {
            Output("pcmsent Broken! (2)\n");
            pcmsent = (pcmsent>>1)<<1;
        }
    }
    if (pcmstart > pcmend)
    {
        if ((pcmsent < pcmstart) && (pcmsent > pcmend))
        {
            Output("pcmsent Broken! (3) start=%04x sent=%04x end=%04x\n", pcmstart, pcmsent, pcmend);
            pcmsent = (pcmsent>>1)<<1;
        }
    }
}
#endif

void audio_close(void)
{
#ifdef _WIN32_WCE
  if (wince_audio_dev != NULL)
  {
    waveOutClose(wince_audio_dev);
    wince_audio_dev = NULL;
  }
#endif
}

int audio_fillblock(ogg_int16_t **pos)
{
    *pos = &pcm[pcmend];
    if (pcmstart == pcmend)
    {
        /* Empty buffer */
        //Output("Sound buffer empty, requesting %d samples\n", PCM_SIZE/4);
        return PCM_SIZE/4;
    }
    else if ((pcmstart - pcmend) > (PCM_SIZE/4))
    {
        /* [***.......#########*****] */
        /*     ^end   ^start   ^sent */
        //Output("Sound buffer wrapped, requesting %d samples\n", PCM_SIZE/4);
        return PCM_SIZE/4;
    }
    else if ((pcmstart == 0) && (pcmend >= 3*PCM_SIZE/4))
    {
        return 0;
    }
    else if (pcmend > pcmstart)
    {
        int max;
        /* [...##########******.....] */
        /*     ^start    ^sent ^end   */

        max = PCM_SIZE-pcmend;
        if (max > PCM_SIZE/4)
            max = PCM_SIZE/4;
        //Output("Sound buffer unwrapped, requesting %d samples\n", max);
        return max;
    }
    //Output("Sound buffer full\n");
    return 0;
}

double audio_time()
{
    int extratime;

    extratime = clock()-audio_sendtime;
    if (!vorbis_p)
      return (double)extratime/CLOCKS_PER_SEC;
    return ((double)audio_samples/vi.channels/vi.rate +
            (double)extratime/CLOCKS_PER_SEC);
}

void audio_filled(int samples)
{
  int fill;
  //Output("Sound buffer filled with %d samples\n", samples);
  pcmend += samples;
  if (pcmend == PCM_SIZE)
    pcmend = 0;
  AUDIO_CHECK();

  fill = pcmend - pcmstart;
  if (fill < 0)
    fill = PCM_SIZE - fill;
  audiobuf_ready = (fill >= PCM_SIZE/2);
}

void audio_write_nonblocking(void)
{
#ifdef _WIN32_WCE
  int      len;
  MMRESULT res;
  DWORD    flags1;
  DWORD    flags2;

  if (wince_audio_dev != NULL)
  {
    flags1 = whdr1.dwFlags;
    if ((flags1 & WHDR_DONE) != 0)
    {
      pcmstart += whdr1.dwBufferLength>>1;
      if (pcmstart >= PCM_SIZE)
        pcmstart -= PCM_SIZE;
      whdr1.dwBufferLength = 0;
      AUDIO_CHECK();
    }
    flags2 = whdr2.dwFlags;
    if ((flags1 & WHDR_DONE) != 0)
    {
      pcmstart += whdr2.dwBufferLength>>1;
      if (pcmstart >= PCM_SIZE)
        pcmstart -= PCM_SIZE;
      whdr2.dwBufferLength = 0;
      AUDIO_CHECK();
    }
    len = (pcmend-pcmsent);
    if (len < 0)
      len = PCM_SIZE-pcmsent;
    if (len == 0)
    {
        /* Nothing to do, as everything sent */
    }
    else if ((flags1 & WHDR_DONE) != 0)
    {
      whdr1.lpData         = (LPSTR)&pcm[pcmsent];
      whdr1.dwBufferLength = len*2; /* Bytes */
      whdr1.dwFlags        = 0;
      res = waveOutPrepareHeader(wince_audio_dev, &whdr1, sizeof(whdr1));
      res = waveOutWrite(wince_audio_dev, &whdr1, sizeof(whdr1));
      audio_samples += len;
      audio_sendtime = clock();
      //Output("Sending %d samples\n", len);
      pcmsent += len;
      if (pcmsent == PCM_SIZE)
        pcmsent = 0;
      AUDIO_CHECK();
    }
    else if ((flags2 & WHDR_DONE) != 0)
    {
      whdr2.lpData         = (LPSTR)&pcm[pcmsent];
      whdr2.dwBufferLength = len*2; /* Bytes */
      whdr2.dwFlags        = 0;
      res = waveOutPrepareHeader(wince_audio_dev, &whdr2, sizeof(whdr2));
      res = waveOutWrite(wince_audio_dev, &whdr2, sizeof(whdr2));
      audio_samples += len;
      audio_sendtime = clock();
      //Output("Sending %d samples\n", len);
      pcmsent += len;
      if (pcmsent == PCM_SIZE)
        pcmsent = 0;
      AUDIO_CHECK();
    }
  }
#endif
  audiobuf_ready = 0;
}


int          raw;
short       *rgb_frame;

FILE* outfile;

void display_video(void *base)
{
    int w, h;

  yuv_buffer yuv;
  theora_decode_YUVout(&td,&yuv);

  w = yuv.y_width;
  if (w > screen_width)
      w = screen_width;
  h = yuv.y_height;
  if (w > screen_height)
      w = screen_height;
  yuv420_2_rgb565(base,
                  yuv.y,
                  yuv.u,
                  yuv.v,
                  h,
                  w,
                  yuv.y_stride,
                  yuv.uv_stride,
                  dither++);
}

#if 0
int got_sigint=0;
static void sigint_handler (int signal) {
  got_sigint = 1;
}
#endif

/* this is a nop in the current implementation. we could
   open a file here or something if so moved. */
static void open_video(void){
  return;
}

/* write out the planar YUV frame, uncropped */
static void video_write(void){
  yuv_buffer yuv;
  theora_decode_YUVout(&td,&yuv);

#if 1
  if (rgb_frame == NULL)
  {
    rgb_frame = malloc(yuv.y_height*yuv.y_width*sizeof(short));
    if (rgb_frame == NULL)
      Output("Failed to malloc output rgb buffer!\n");
  }
  if (rgb_frame != NULL)
  {
      yuv420_2_rgb565(rgb_frame,
                      yuv.y,
                      yuv.u,
                      yuv.v,
                      yuv.y_height,
                      yuv.y_width,
                      yuv.y_stride,
                      yuv.uv_stride,
                      dither++);
    Output("Saving RGB Frame\n");
    fwrite(rgb_frame, 1, yuv.y_height*yuv.y_width*sizeof(short), outfile);
  }

#else
  if(outfile){
    int i;
    Output("Saving Frame\n");
    if(!raw)
      fprintf(outfile, "FRAME\n");
    for(i=0;i<yuv.y_height;i++)
      fwrite(yuv.y+yuv.y_stride*i, 1, yuv.y_width, outfile);
    for(i=0;i<yuv.uv_height;i++)
      fwrite(yuv.u+yuv.uv_stride*i, 1, yuv.uv_width, outfile);
    for(i=0;i<yuv.uv_height;i++)
      fwrite(yuv.v+yuv.uv_stride*i, 1, yuv.uv_width, outfile);
  }
#endif
}

static void do_crc(unsigned char *in, int width, int *crc_ref)
{
    int crc = *crc_ref;

    while (--width > 0)
    {
      int i = *in++;

      if (--width > 0)
          i |= (*in++)<<8;

      crc = ((crc<<1) | (crc>>15)) ^ i;
      crc &= 65535;
    }

    *crc_ref = crc;
}

static void video_crc(int *crcref){
  int i;

  yuv_buffer yuv;
  theora_decode_YUVout(&td,&yuv);

  for(i=0;i<yuv.y_height;i++)
    do_crc(yuv.y+yuv.y_stride*i, yuv.y_width, crcref);
  for(i=0;i<yuv.uv_height;i++)
    do_crc(yuv.u+yuv.uv_stride*i, yuv.uv_width, crcref);
  for(i=0;i<yuv.uv_height;i++)
    do_crc(yuv.v+yuv.uv_stride*i, yuv.uv_width, crcref);

  Output("Cumulative CRC = %04x\n", *crcref);
}

/* dump the theora comment header */
static int dump_comments(theora_comment *tc){
  int i, len;
  char *value;

  Output("Encoded by %s\n",tc->vendor);
  if(tc->comments){
    Output("theora comment header:\n");
    for(i=0;i<tc->comments;i++){
      if(tc->user_comments[i]){
        len=tc->comment_lengths[i];
        value=malloc(len+1);
        memcpy(value,tc->user_comments[i],len);
        value[len]='\0';
        Output("\t%s\n", value);
        free(value);
      }
    }
  }
  return(0);
}



/* helper: push a page into the appropriate steam */
/* this can be done blindly; a stream won't accept a page
                that doesn't belong to it */
static int queue_page(ogg_page *page){
  if(theora_p)ogg_stream_pagein(&to,page);
  if(vorbis_p)ogg_stream_pagein(&vo,page);
  return 0;
}

static void syntax(void){
  Output( "Usage: testtheorarm <file.ogv> [<outfile>] [maxframes]\n"
          "\n"
  );
  exit(1);
}

int main(int argc,char *const *argv){

  ogg_packet op;

  void *framebuffer = NULL;
  int max_frames = 1;
  int frames     = 0;
  int crc        = 0;
  int refcrc     = -1;
#ifdef POST_PROCESS
  int pp_level_max;
  int pp_level;
  int pp_inc;
#endif

  clock_t start_clock, end_clock;
  FILE *infile   = stdin;
  outfile        = NULL;
  rgb_frame      = NULL;

  /* Process option arguments. */
  if (argc < 2)
    syntax();

  infile=fopen(argv[1],"rb");
  if(infile==NULL){
    Output("Unable to open '%s' for extraction.\n", argv[1]);
    syntax();
  }

  playback = Playback_DecodeOnly;
  if (argc >= 3)
  {
    if (argv[2][0] == '-')
    {
      sscanf(&argv[2][1], "%x", &refcrc);
    }
    else if (argv[2][0] == '+')
    {
      playback = Playback_Synced; /* Really crap attempt at syncing */
    }
    else if (argv[2][0] == '*')
    {
      playback = Playback_Unsynced; /* As fast as we can! */;
    }
    else
    {
      outfile=fopen(argv[2],"wb");
      if(outfile==NULL){
        Output("Unable to open output file '%s'\n", argv[2]);
        exit(1);
      }
    }
  }

  if (argc >= 4)
  {
    max_frames=atoi(argv[3]);
  }

  theora_p = 0;
  vorbis_p = 0;
  stateflag = 0;
  videobuf_ready=0;
  videobuf_granulepos=-1;
  videobuf_time=0;
  audiobuf_ready=0;
  audiobuf_granulepos=0; /* time position of last sample */
  raw=0;

  /*
     Ok, Ogg parsing. The idea here is we have a bitstream
     that is made up of Ogg pages. The libogg sync layer will
     find them for us. There may be pages from several logical
     streams interleaved; we find the first theora stream and
     ignore any others.

     Then we pass the pages for our stream to the libogg stream
     layer which assembles our original set of packets out of
     them. It's the packets that libtheora actually knows how
     to handle.
  */

  /* start up Ogg stream synchronization layer */
  ogg_sync_init(&oy);

  /* init supporting Vorbis structures needed in header parsing */
  vorbis_info_init(&vi);
  vorbis_comment_init(&vc);

  /* init supporting Theora structures needed in header parsing */
  theora_comment_init(&tc);
  theora_info_init(&ti);

  memset(&op, 0, sizeof(op));

  /* Ogg file open; parse the headers */

  /* Vorbis and Theora both depend on some initial header packets
     for decoder setup and initialization. We retrieve these first
     before entering the main decode loop. */

  /* Only interested in Theora streams */
  while(!stateflag){
    int ret=buffer_data(infile,&oy);
    if(ret==0)break;
    while(ogg_sync_pageout(&oy,&og)>0){
      ogg_stream_state test;

      /* is this a mandated initial header? If not, stop parsing */
      if(!ogg_page_bos(&og)){
        /* don't leak the page; get it into the appropriate stream */
        queue_page(&og);
        stateflag=1;
        break;
      }

      ogg_stream_init(&test,ogg_page_serialno(&og));
      ogg_stream_pagein(&test,&og);
      ogg_stream_packetout(&test,&op);

      /* identify the codec: try theora */
      if(!theora_p && theora_decode_header(&ti,&tc,&op)>=0){
        /* it is theora -- save this stream state */
        memcpy(&to,&test,sizeof(test));
        theora_p=1;
      }else if(!vorbis_p && vorbis_dsp_headerin(&vi,&vc,&op)>=0){
        /* it is vorbis */
        memcpy(&vo,&test,sizeof(test));
        vorbis_p=1;
      }else{
        /* whatever it is, we don't care about it */
        ogg_stream_clear(&test);
      }
    }
    /* fall through to non-bos page parsing */
  }

  /* we're expecting more header packets. */
  while((theora_p && theora_p<3) || (vorbis_p && vorbis_p<3)){
    int ret;

    /* look for further theora headers */
    while(theora_p && (theora_p<3) && (ret=ogg_stream_packetout(&to,&op))){
      if(ret<0){
        Output("Error parsing Theora stream headers; corrupt stream?\n");
        exit(1);
      }
      if(theora_decode_header(&ti,&tc,&op)){
        Output("Error parsing Theora stream headers; corrupt stream?\n");
        exit(1);
      }
      theora_p++;
    }

    /* look for more vorbis header packets */
    while(vorbis_p && (vorbis_p<3) && (ret=ogg_stream_packetout(&vo,&op))){
      if(ret<0){
        fprintf(stderr,"Error parsing Vorbis stream headers; corrupt stream?\n");
        exit(1);
      }
      if(vorbis_dsp_headerin(&vi,&vc,&op)){
        fprintf(stderr,"Error parsing Vorbis stream headers; corrupt stream?\n");
        exit(1);
      }
      vorbis_p++;
    }

    /* The header pages/packets will arrive before anything else we
       care about, or the stream is not obeying spec */

    if(ogg_sync_pageout(&oy,&og)>0){
      queue_page(&og); /* demux into the appropriate stream */
    }else{
      int ret=buffer_data(infile,&oy); /* someone needs more data */
      if(ret==0){
        Output("End of file while searching for codec headers.\n");
        exit(1);
      }
    }
  }

  /* and now we have it all.  initialize decoders */
  if(theora_p){
    theora_decode_init(&td,&ti);
    Output("Ogg logical stream %lx is Theora %dx%d %.02f fps video\n"
           "Encoded frame content is %dx%d with %dx%d offset\n",
           to.serialno,ti.width,ti.height,
           (double)ti.fps_numerator/ti.fps_denominator,
           ti.frame_width,ti.frame_height,ti.offset_x,ti.offset_y);
    if (playback != Playback_DecodeOnly)
      framebuffer = open_screen(ti.width, ti.height);
    dump_comments(&tc);
#ifdef POST_PROCESS
    theora_control(&td,TH_DECCTL_GET_PPLEVEL_MAX,&pp_level_max,
     sizeof(pp_level_max));
    pp_level=pp_level_max;
    theora_control(&td,TH_DECCTL_SET_PPLEVEL,&pp_level,sizeof(pp_level));
    pp_inc=0;
#endif
  }else{
    /* tear down the partial theora setup */
    theora_info_clear(&ti);
    theora_comment_clear(&tc);
  }

  if(vorbis_p){
    vorbis_dsp_init(&vd,&vi);
    Output("Ogg logical stream %lx is Vorbis %d channel %ld Hz audio.\n",
           vo.serialno,vi.channels,vi.rate);
  }else{
    /* tear down the partial vorbis setup */
    vorbis_info_clear(&vi);
    vorbis_comment_clear(&vc);
  }

  /* open audio */
  if(vorbis_p)open_audio();

  /* open video */
  if(theora_p)open_video();

  if(!raw && outfile){
    Output("YUV4MPEG2 W%d H%d F%d:%d I%c A%d:%d\n",
           ti.width,ti.height,ti.fps_numerator,ti.fps_denominator,'p',
           ti.aspect_numerator,ti.aspect_denominator);
  }

#if 0
  /* install signal handler */
  signal (SIGINT, sigint_handler);
#endif

  start_clock = clock();

  /* Finally the main decode loop.

     It's one Theora packet per frame, so this is pretty
     straightforward if we're not trying to maintain sync
     with other multiplexed streams.

     the videobuf_ready flag is used to maintain the input
     buffer in the libogg stream state. If there's no output
     frame available at the end of the decode step, we must
     need more input data. We could simplify this by just
     using the return code on ogg_page_packetout(), but the
     flag system extends easily to the case were you care
     about more than one multiplexed stream (like with audio
     playback). In that case, just maintain a flag for each
     decoder you care about, and pull data when any one of
     them stalls.

     videobuf_time holds the presentation time of the currently
     buffered video frame. We ignore this value.
  */

  stateflag=0; /* playback has not begun */
  /* queue any remaining pages from data we buffered but that did not
      contain headers */
  while(ogg_sync_pageout(&oy,&og)>0){
    queue_page(&og);
  }

  //if(fps_only){
  //  ftime(&start);
  //  ftime(&last);
  //}

  //while(!got_sigint){
  while(max_frames > 0){

    /* we want a video and audio frame ready to go at all times.  If
       we have to buffer incoming, buffer the compressed data (ie, let
       ogg do the buffering) */
    while(vorbis_p && !audiobuf_ready){
      int ret;
      ogg_int16_t *fillbase;

      /* if there's pending, decoded audio, grab it */
      ret=audio_fillblock(&fillbase);
      if (ret==0)
          break;
      ret=vorbis_dsp_pcmout(&vd,fillbase,ret/vi.channels);
      if(ret>0){
        vorbis_dsp_read(&vd,ret);
        if(vd.granulepos>=0)
          audiobuf_granulepos=vd.granulepos;
        else
          audiobuf_granulepos+=ret;
        ret *= vi.channels;
        audio_filled(ret);

      }else{

        /* no pending audio; is there a pending packet to decode? */
        if(ogg_stream_packetout(&vo,&op)>0){
          vorbis_dsp_synthesis(&vd,&op,1);//==0) /* test for success! */
            //vorbis_synthesis_blockin(&vd,&vb);
        }else   /* we need more data; break out to suck in another page */
          break;
      }
    }

    while(theora_p && !videobuf_ready){
      /* theora is one in, one out... */
      if(ogg_stream_packetout(&to,&op)>0){
        int ret;

#ifdef POST_PROCESS
        if(pp_inc){
          pp_level+=pp_inc;
          theora_control(&td,TH_DECCTL_SET_PPLEVEL,&pp_level,
           sizeof(pp_level));
          pp_inc=0;
        }
#endif

        ret=theora_decode_packetin(&td,&op);
        if (ret<0)
        {
          Output("Decode returned %d\n",ret);
        }
        videobuf_granulepos=td.granulepos;
        videobuf_time=theora_granule_time(&td,videobuf_granulepos);
        videobuf_ready=1;
        frames++;
        max_frames--;
        //if(fps_only)
        //  ftime(&after);

      }else
        break;
    }

    if(!videobuf_ready && !audiobuf_ready && feof(infile))break;

    if((theora_p && !videobuf_ready) || (vorbis_p && !audiobuf_ready)){
      /* no data yet for somebody.  Grab another page */
      buffer_data(infile,&oy);
      while(ogg_sync_pageout(&oy,&og)>0){
        queue_page(&og);
      }
    }

    /* If playback has begun, top audio buffer off immediately. */
    if(stateflag) audio_write_nonblocking();

    /* dumpvideo frame, and get new one */
    if(stateflag && videobuf_ready &&
       ((playback!=Playback_Synced) || (videobuf_time < audio_time())))
    {
      if (outfile)
      {
        video_write();
      }
      else if (refcrc != -1)
      {
        video_crc(&crc);
      }
      if ((videobuf_ready) && (framebuffer != NULL))
      {
        display_video(framebuffer);
      }
      videobuf_ready=0;
    }

    /* if our buffers either don't exist or are ready to go,
       we can begin playback */
    if((!theora_p || videobuf_ready) &&
       (!vorbis_p || audiobuf_ready))stateflag=1;
    /* same if we've run out of input */
    if(feof(infile))stateflag=1;

  }

  end_clock = clock();

  /* end of decoder loop -- close everything */
  audio_close();

  if(vorbis_p){
    ogg_stream_clear(&vo);
    vorbis_dsp_clear(&vd);
    vorbis_comment_clear(&vc);
    vorbis_info_clear(&vi);
  }
  if(theora_p){
    ogg_stream_clear(&to);
    theora_clear(&td);
    theora_comment_clear(&tc);
    theora_info_clear(&ti);
  }
  ogg_sync_clear(&oy);

  free(rgb_frame);

  if(infile && infile!=stdin)fclose(infile);
  if(outfile) fclose(outfile);
  if(refcrc != -1)
  {
    if (crc == refcrc)
      Output("CRC good!\n");
    else
      Output("CRC bad! %04x, but expected %04x\n", crc, refcrc);
  }

  end_clock -= start_clock;
  Output("\n\n%d frames in %f ms  (%f fps)\n",
         frames, ((float)end_clock)/CLOCKS_PER_SEC,
         ((float)frames)*CLOCKS_PER_SEC/end_clock);
  Output("\nDone.\n");

  return(0);

}

typedef struct
{
    int    argc;
    char **argv;
}
thread_args;

DWORD WINAPI thread_starter(void *args_)
{
    thread_args *args = (thread_args *)args_;

    return main(args->argc, args->argv);
}

unsigned __int64 FILETIME_to_uint64(FILETIME ft)
{
    unsigned __int64 ret;
    ret = ft.dwHighDateTime;
    ret <<= 32;
    ret |= ft.dwLowDateTime;
    return ret;
}

int profile(int argc, char *argv[])
{
    HANDLE      thread;
    thread_args args;
    DWORD       threadId;
    DWORD       exitcode = 1;
    //FILETIME    create, exit, kernel, user;
    int         total_time = 0;
    unsigned __int64 oldTime = 0;
    CONTEXT     context;
    int        *profile;
    int         bin;
    int         tickcount = 0;

#define PROFILE_START (0x11000)
#define PROFILE_STOP  (0x39900)
#define PROFILE_SIZE  (PROFILE_STOP-PROFILE_START)
#define PROFILE_QUANT (4)

    args.argc = argc;
    args.argv = argv;

    profile = calloc(PROFILE_SIZE>>PROFILE_QUANT, 4);
    if (profile == NULL)
    {
        Output("Failed to allocate profile space!\n");
        return 1;
    }

    thread = CreateThread(NULL,
                          0,
                          (LPTHREAD_START_ROUTINE)&thread_starter,
                          &args,
                          0,
                          &threadId);

    while (1 == 1)
    {
        BOOL  done;

        done = GetExitCodeThread(thread, &exitcode);
        if ((!done) || (exitcode != STILL_ACTIVE))
            break;
#if 0
        done = GetThreadTimes(thread,
                              &create,
                              &exit,
                              &kernel,
                              &user);
        if (done)
        {
            unsigned __int64 t64;

            t64 = FILETIME_to_uint64(user) + FILETIME_to_uint64(kernel);
            oldTime = (t64-oldTime)/10000;
            total_time += (int)oldTime;
            oldTime = t64;
        }
#endif
        context.ContextFlags = CONTEXT_FULL;
        done = GetThreadContext(thread, &context);

        if (done)
        {
            bin = (context.Pc & 0x7FFFFFFF)>>(PROFILE_QUANT+2);
            if (bin != 0)
                bin -= PROFILE_START>>(PROFILE_QUANT+2);
            if ((bin < 0) || (bin >= PROFILE_SIZE))
            {
                /* Try again, but with r14 */
                bin = (context.Lr & 0x7FFFFFFF)>>(PROFILE_QUANT+2);
                if (bin != 0)
                    bin -= PROFILE_START>>(PROFILE_QUANT+2);
            }
            if ((bin < 0) || (bin >= PROFILE_SIZE))
            {
                bin = 0;
                //Output("PC=%x r14=%x PSR=%x\n", context.Pc, context.Lr, context.Psr);
            }
            profile[bin]++;
            tickcount++;
        }

        Sleep(1);
    }

    Output("Profiler shutting down: %x\n", total_time);

    {
        FILE *file;

        file = fopen("\\Storage Card\\profile.out", "wb");
        if (file != NULL)
        {
            int i;

            fputc('P',file);
            fputc('R',file);
            fputc('O',file);
            fputc('F',file);
            i = 1; /* File version */
            fwrite(&i, sizeof(int), 1, file);
            i = 1000; /* Frequency */
            fwrite(&i, sizeof(int), 1, file);
            fwrite(&tickcount, sizeof(int), 1, file);
            i = 1; /* Sections */
            fwrite(&i, sizeof(int), 1, file);
            i = PROFILE_QUANT; /* Granularity */
            fwrite(&i, sizeof(int), 1, file);
            i = PROFILE_START; /* Start */
            fwrite(&i, sizeof(int), 1, file);
            i = PROFILE_SIZE>>PROFILE_QUANT; /* Size */
            fwrite(&i, sizeof(int), 1, file);
            fwrite(profile, sizeof(int), PROFILE_SIZE>>PROFILE_QUANT, file);
            fclose(file);
        }
    }

    return exitcode;
}


#define TESTFILE 0

int WinMain(HINSTANCE h,HINSTANCE i,LPWSTR l,int n)
{
    int ret;
    char *argv0[] = { "testtheorarm",
                     "\\Storage Card\\theorarm\\matrix-55.ogg",
                     //"-662c", // No PP
                     "-47e5", // Full PP
                     "10",
                     NULL
                   };
    char *argv1[]= { "testtheorarm",
                     "\\Storage Card\\theorarm\\matrix-300-varAQ.ogg",
                     //"\\Storage Card\\theorarm\\matrix-300-varAQ.out",
                     //"-90d6", // No PP
                     "-3ddf", // Full PP
                     "10",
                     NULL
                   };
    char *argv0v[] = { "testtheorarm",
                     "\\Storage Card\\theorarm\\matrix-55.ogg",
                     "+",
                     "100",
                     NULL
                   };
    char *argv1v[]= { "testtheorarm",
                     "\\Storage Card\\theorarm\\matrix-300-varAQ.ogg",
                     //"\\Storage Card\\theorarm\\matrix-300-varAQ.out",
                     "+",
                     "100",
                     NULL
                   };
    char *argv2v[]= { "testtheorarm",
                     //"\\Storage Card\\theorarm\\my-matrix.ogg",
                     "\\my-matrix2.ogg",
                     "*",
                     "100000000",
                     NULL
                   };

    ret = main(4, argv0);
    if (ret == 0)
        ret = main(4, argv1);
    if (ret == 0)
        ret = main(4, argv0v);
    if (ret == 0)
        ret = main(4, argv1v);
    if (ret == 0)
        ret = profile(4, argv2v);
    return ret;
}

int wmain(int argc,char *const *argv)
{
    return WinMain(0,0,0,0);
}
